/*
Voici le programme c de la recherche de zeros. La fonction qui trouve
les racines est

        roots(GEN p, long l)

qui renvoie le vecteur des racines avec une erreur absolue garantie de
BASE^(-l), ou BASE=2^{BITS_IN_LONG}.

*/

#include "genpari.h"

#define INFINITY 100000
#define NEWTON_MAX 10

GEN gunr, gzeror, u;
long KARASQUARE_LIMIT, COOK_SQUARE_LIMIT, Lmax, step4;
double *radius;

/********************************************************************/
/********************************************************************/
/**                                                                **/
/**         ARITHMETIQUE RAPIDE UTILISEE POUR GRAEFFE              **/
/**                                                                **/
/********************************************************************/
/********************************************************************/


/* using karatsuba, fast product of integers x,y */
GEN 
quickmulii(GEN a,GEN b){
    long na=lgef(a)-2,nb=lgef(b)-2,n0,nn0,nnb,n1,i,ltop=avma,lbot;
    GEN a0,aa,b0,bb,c0,c1,c2,c;
    
    if (na<nb) return quickmulii(b,a);
    if (na<30) return mulii(a,b);

    n1=(na>>1); n0=na-n1; nn0=n0+2;
    setlgef(a,n1+2);
    a0=cgetg(nn0,1); a0[1]=evalsigne(signe(a))+evallgef(nn0);
    aa=a+n1; for (i=2; i<nn0; i++) a0[i]=aa[i];

    if (nb>n0){
        nnb=nb-n0; setlgef(b,nnb+2);
        b0=cgetg(nn0,1); b0[1]=evalsigne(signe(b))+evallgef(nn0);
        bb=b+nnb; for (i=2; i<nn0; i++) b0[i]=bb[i];
        c2=quickmulii(a,b); c0=quickmulii(a0,b0);
        c1=quickmulii(addii(a,a0),addii(b,b0)); c1=subii(c1,addii(c0,c2));
        c=addii(shifti(c2,(n0<<(TWOPOTBITS_IN_LONG+1))),
                shifti(c1,(n0<<TWOPOTBITS_IN_LONG)));
        setlgef(b,nb+2);
    }
    else{
        c0=quickmulii(a0,b); c1=quickmulii(a,b);
        c=shifti(c1,(n0<<TWOPOTBITS_IN_LONG));
    }
    setlgef(a,na+2); 
    lbot=avma;
    return gerepile(ltop,lbot,addii(c,c0));
}
    

/* fast product of x,y which must be integer or complex of integer */
GEN quickmulcc(GEN x, GEN y){

    long tx=typ(x),ty=typ(y),l,tetpil;
    GEN p1,p2,p3,p4,p5,z;

    if (tx==1){
        if (ty==1) return quickmulii(x,y);
        if (ty==6){
            z=cgetg(3,6); 
            z[1]=(long) quickmulii(x,(GEN) y[1]);
            z[2]=(long) quickmulii(x,(GEN) y[2]);
            return z;
        }
    }
    if (tx==6){
        if (ty==1){
            z=cgetg(3,6);
            z[1]=(long) quickmulii((GEN)x[1],y);
            z[2]=(long) quickmulii((GEN)x[2],y);
            return z;
        }
        if (ty==6){
            z=cgetg(3,6);l=avma;
            p1=quickmulii((GEN)x[1],(GEN)y[1]);
            p2=quickmulii((GEN)x[2],(GEN)y[2]);
            p3=addii((GEN)x[1],(GEN)x[2]);
            p4=addii((GEN)y[1],(GEN)y[2]);
            p5=quickmulii(p3,p4);
            p3=addii(p1,p2);
            tetpil=avma;
            z[2]=(long) subii(p5,p3);
            z[1]=(long) subii(p1,p2);
            z[1]=lpile(l,tetpil,(GEN)z[1]);
            return z;
        }
    }
    err(talker,"bug in quickmulcc");return gnil;
}

void set_karasquare_limit(long bitprec){
    
    if (bitprec<600) { KARASQUARE_LIMIT=8; COOK_SQUARE_LIMIT=400; return; }
    if (bitprec<2000) { KARASQUARE_LIMIT=4; COOK_SQUARE_LIMIT=200; return; }
    if (bitprec<3000) { KARASQUARE_LIMIT=4; COOK_SQUARE_LIMIT=125; return; }
    if (bitprec<5000) { KARASQUARE_LIMIT=2; COOK_SQUARE_LIMIT=75; return; }
    KARASQUARE_LIMIT=1; COOK_SQUARE_LIMIT=50;
}

/* the pari library does not have specific procedure for the square of 
polynomials. This one is twice faster than gmul */
GEN 
mysquare(GEN p){
    GEN s,aux1,aux2;
    long i,j,n=lgef(p)-3,nn,ltop,lbot;

    if (n==-1) return gcopy(p);
    nn=n<<1;
    s=cgetg(nn+3,10);
    s[1]=evalsigne(1)+evalvarn(varn(p))+evallgef(nn+3);
    for (i=0; i<=n; i++){
        aux1=gzero; 
        ltop=avma;
        for (j=0; j<(i+1)>>1; j++){
            aux2=quickmulcc((GEN) p[j+2],(GEN)p[i-j+2]);
            aux1=gadd(aux1,aux2);
        }
        if (i%2==1){ lbot=avma; s[i+2]=lpile(ltop,lbot,gshift(aux1,1)); }
        else{
            aux1=gshift(aux1,1);
            aux2=quickmulcc((GEN)p[2+(i>>1)],(GEN)p[2+(i>>1)]);
            lbot=avma; s[i+2]=lpile(ltop,lbot,gadd(aux1,aux2));
        }
    }
    for (i=n+1; i<=nn; i++){
        aux1=gzero; ltop=avma;
        for (j=i-n; j<(i+1)>>1; j++){
            aux2=quickmulcc((GEN)p[j+2],(GEN)p[i-j+2]);
            aux1=gadd(aux1,aux2);
        }
        if (i%2==1){ lbot=avma; s[i+2]=lpile(ltop,lbot,gshift(aux1,1));}
        else{
            aux1=gshift(aux1,1);
            aux2=quickmulcc((GEN)p[2+(i>>1)],(GEN)p[2+(i>>1)]);
            lbot=avma; s[i+2]=lpile(ltop,lbot,gadd(aux1,aux2));
        }       
    }
    return s;
}

GEN 
karasquare(GEN p){
    GEN p1,s0,s1,s2,aux;
    long n=lgef(p)-3,n0,n1,i,var,nn0,ltop,lbot;
    
    if (n<=KARASQUARE_LIMIT) return mysquare(p);
    ltop=avma;
    var=evalsigne(1)+evalvarn(varn(p)); n0=n>>1; n1=n-n0-1;
    setlgef(p,n0+3);  /* hack to have the first half of p */
    s0=karasquare(p);
    p1=cgetg(n1+3,10); p1[1]=var+evallgef(n1+3);
    for (i=2; i<=n1+2; i++) p1[i]=p[1+i+n0];
    s2=karasquare(p1);
    s1=karasquare(gadd(p,p1));
    s1=gsub(s1,gadd(s0,s2));
    nn0=n0<<1;
    aux=cgetg((n<<1)+3,10); aux[1]=var+evallgef(2*n+3);
    var=lgef(s0);
    for (i=2; i<var; i++) aux[i]=s0[i];
    for (i=var; i<=nn0+2; i++) aux[i]=zero;
    var=lgef(s2);
    for (i=2; i<var; i++) aux[2+i+nn0]=s2[i];
    for (i=var-2; i<=(n1<<1); i++) aux[4+i+nn0]=zero;
    aux[3+nn0]=zero;
    for (i=3; i<=lgef(s1); i++)
        aux[i+n0]=ladd((GEN) aux[i+n0],(GEN) s1[i-1]);
    setlgef(p,n+3); /* recover all the polynomial p */
    lbot=avma; return gerepile(ltop,lbot,gcopy(aux));
}


GEN 
cook_square(GEN p){
    GEN p0,p1,p2,p3,q,aux0,aux1,r,aux,plus,moins;
    long n=lgef(p)-3,n0,n3,i,j,ltop=avma,lbot,var;
    
    if (n<=COOK_SQUARE_LIMIT) return karasquare(p);
    
    n0=(n+1)/4; n3=n+1-3*n0;
    p0=cgetg(n0+2,10); p1=cgetg(n0+2,10); p2=cgetg(n0+2,10); 
    p3=cgetg(n3+2,10);
    var=evalsigne(1)+evalvarn(varn(p));
    p0[1]=p1[1]=p2[1]=var+evallgef(n0+2); p3[1]=var+evallgef(n3+2);
    
    for (i=0; i<n0; i++){
        p0[i+2]=p[i+2]; p1[i+2]=p[i+n0+2]; p2[i+2]=p[i+2*n0+2];
    }
    for (i=0; i<n3; i++) p3[i+2]=p[i+3*n0+2];

    q=cgetg(8,17); q=q+4;

    q[0]=(long) p0;
    aux0=gadd(p0,p2); aux1=gadd(p1,p3);
    q[-1]=lsub(aux0,aux1); q[1]=ladd(aux0,aux1);
    aux0=gadd(p0,gmulgs(p2,4)); aux1=gmulgs(gadd(p1,gmulgs(p3,4)),2);
    q[-2]=lsub(aux0,aux1); q[2]=ladd(aux0,aux1);
    aux0=gadd(p0,gmulgs(p2,9)); aux1=gmulgs(gadd(p1,gmulgs(p3,9)),3);
    q[-3]=lsub(aux0,aux1); q[3]=ladd(aux0,aux1);
    for (i=-3; i<=3; i++) q[i]=(long) cook_square((GEN)q[i]);
    r=cgetg(8,17); r++;
    plus=cgetg(4,17); moins=cgetg(4,17);
    for (i=1; i<=3; i++){
        plus[i]=ladd((GEN)q[-i],(GEN)q[i]);
        moins[i]=lsub((GEN)q[-i],(GEN)q[i]);
    }
    r[0]=q[0];
    r[1]=ldivgs(
                gsub(
                     gsub(gmulgs((GEN)moins[2],9),(GEN)moins[3]),
                     gmulgs((GEN)moins[1],45)),
                60);
    r[2]=ldivgs(
                gadd(
                     gadd(gmulgs((GEN)plus[1],270),gmulgs((GEN)q[0],-490)),
                     gadd(gmulgs((GEN)plus[2],-27),gmulgs((GEN)plus[3],2))),
                360);
    r[3]=ldivgs(
                gadd(
                     gadd(gmulgs((GEN)moins[1],13),gmulgs((GEN)moins[2],-8)),
                     (GEN)moins[3]),
                48);
    r[4]=ldivgs(
                gadd(
                     gadd(gmulgs((GEN)q[0],56),gmulgs((GEN)plus[1],-39)),
                     gsub(gmulgs((GEN)plus[2],12),(GEN)plus[3])),
                144);
    r[5]=ldivgs(
                gsub(
                     gadd(gmulgs((GEN)moins[1],-5),gmulgs((GEN)moins[2],4)),
                     (GEN)moins[3]),
                240);
    r[6]=ldivgs(
                gadd(
                     gadd(gmulgs((GEN)q[0],-20),gmulgs((GEN)plus[1],15)),
                     gadd(gmulgs((GEN)plus[2],-6),(GEN)plus[3])),
                720);
    q=cgetg(2*n+3,10); q[1]=var+evallgef(2*n+3);
    for (i=0; i<=2*n; i++) q[i+2]=zero;
    for (i=0; i<=6; i++){
        aux=(GEN) r[i];
        for (j=0; j<=lgef(aux)-3; j++){
            q[n0*i+j+2]=ladd((GEN)q[n0*i+j+2],(GEN)aux[j+2]);
        }
    }
    lbot=avma; return gerepile(ltop,lbot,gcopy(q));
}

GEN 
graeffe(GEN p){
    GEN p0,p1,s0,s1,ss1;
    long n=lgef(p)-3,n0,n1,ns1,i,ltop=avma,lbot,auxi;
    
    if (n==0) return p;
    n0=(n>>1); n1=((n-1)>>1);
    auxi=evalsigne(1)+evalvarn(varn(p));
    
    p0=cgetg(n0+3,10); p0[1]=auxi+evallgef(n0+3);
    for (i=0; i<=n0; i++)
        p0[i+2]=p[2+(i<<1)];

    p1=cgetg(n1+3,10); p1[1]=auxi+evallgef(n1+3);
    for (i=0; i<=n1; i++)
        p1[i+2]=p[3+(i<<1)];
    s0=cook_square(p0);
    s1=cook_square(p1);
    ns1=lgef(s1)-3;
    ss1=cgetg(ns1+4,10); /* will contain x*s1 */
    ss1[1]=auxi+evallgef(ns1+4);
    ss1[2]=zero;
    for (i=0; i<=ns1; i++) 
        ss1[3+i]=s1[2+i];
    lbot=avma;
    return gerepile(ltop,lbot,gsub(s0,ss1)); 
}

/********************************************************************/
/********************************************************************/
/**                                                                **/
/**        FACTORISATION SQUAREFREE AVEC LE GCD MODULAIRE          **/
/**                                                                **/
/********************************************************************/
/********************************************************************/


GEN 
maxnorm(GEN p){
    long i,n=lgef(p)-3,ltop=avma,lbot;
    GEN x,m;

    m=gzero;
    for (i=0; i<n; i++){
        x=gabs((GEN)p[2+i],5);
        if (gcmp(x,m)==1) m=x;
    }
    m=gdivent(m,gabs((GEN)p[2+n],5));
    lbot=avma;
    return gerepile(ltop,lbot,gadd(m,gun));
}


/* idem as resmod but working only on C-long types */
long *
resmodp_withlong(long *x, long dx, long *y, long dy, long *dc, long p){
    long dz,i,j,p1,*p4,*c,inv;

    dz=dx-dy; 
    if (dz<0) { 
        *dc=dx;
        c=(long *) malloc((dx+1)*sizeof(long));
        for (i=0; i<=dx; i++) c[i]=x[i];
        return c; }
    p4=(long *) malloc((dz+1)*sizeof(long));
    inv=itos(lift(gdiv(gun,gmodulcp(stoi(y[dy]),stoi(p)))));
    p4[dz]=(inv*x[dx])%p;
    for (i=dx-1; i>=dy; --i){
        p1=x[i];
        for (j=i-dy+1;(j<=i)&&(j<=dz);j++){
            p1-=p4[j]*y[i-j];
            if ((j&15)==0) p1=p1%p;
        }
        p4[i-dy]=((p1%p)*inv)%p;
    }
    c=(long *) malloc(dy*sizeof(long));
    for (i=0; i<dy; i++){
        p1=p4[0]*y[i];
        for (j=1; ((j<=i)&&(j<=dz));j++) {
            p1+=p4[j]*y[i-j];
            if ((j&15)==0) p1=p1%p;
        }

        c[i]=(x[i]-p1)%p;
    }
    i=dy-1; while ((i>=0)&&(c[i]==0)) i--;
    *dc=i;
    free(p4);
    return c;
}

/* x and y in Z[x], return c in Z[x] such that c= x%y mod p */
GEN 
resmodp(GEN x, GEN y, GEN p){
    long dx=lgef(x)-3,dy=lgef(y)-3,dz,i,j,lb,lt,ltop=avma;
    GEN p1,p2,p4,c;

    dz=dx-dy; if (dz<0) return x;
    p4=cgetg(dz+3,10);p4[1]=evalsigne(1)+evallgef(3+dz)+evalvarn(varn(x));
    p4[dz+2]=ldiv(gmodulcp((GEN)x[dx+2],p),(GEN)y[dy+2]);
    for(i=dx-1;i>=dy;--i) {
        p1=gmodulcp((GEN)x[i+2],p); lt=avma;
        for(j=i-dy+1;(j<=i)&&(j<=dz);j++) {
            p2=gmul((GEN)p4[j+2],(GEN)y[i-j+2]);
            p1=gsub(p1,p2);
        } 
        lb=avma; p4[i-dy+2]=lpile(lt,lb,gdiv(p1,(GEN)y[dy+2]));
    }
    p4=lift(p4);
    c=cgetg(dy+2,10); 
    for (i=0; i<dy; i++){
        lt=avma;
        p1=mulii((GEN)p4[2],(GEN)y[i+2]);
        for (j=1;((j<=i) && (j<=dz));j++){
            p1=addii(p1,mulii((GEN)p4[j+2],(GEN)y[i-j+2]));
        }
        lb=avma; c[i+2]=lpile(lt,lb,gmod(gsub((GEN)x[i+2],p1),p)); 
    }
    i=dy-1; while ((i>=0) && (gcmp0((GEN)c[2+i]))) i--;
    if (i==-1) return gzero;
    c[1]=evallgef(i+3)+evalsigne(1)+evalvarn(varn(x));
    lb=avma; return gerepile(ltop,lb,gcopy(c));
}

GEN 
gcdmodp_withlong(GEN x, GEN y, GEN p){
    long *a,*b,*c,da,db,dc,i,pp,ltop=avma,lbot;
    GEN z;

    da=lgef(x)-3; db=lgef(y)-3;
    a=(long *) malloc((da+1)*sizeof(long));
    b=(long *) malloc((db+1)*sizeof(long));
    for (i=0; i<=da; i++) a[i]=itos(gmod((GEN)x[i+2],p));
    while ((da>=0) && (a[da]==0)) da--;
    if (da==-1) {
        free(a); free(b); return gmul(y,gmodulcp(gun,p));
    }
    for (i=0; i<=db; i++) b[i]=itos(gmod((GEN)y[i+2],p));
    while ((db>=0) && (b[db]==0)) db--;
    if (db==-1) {
        free(a); free(b); return gmul(x,gmodulcp(gun,p));
    }
    pp=itos(p);
    while (db>=0){
        c=resmodp_withlong(a,da,b,db,&dc,pp);
        free(a);
        a=b; da=db; b=c; db=dc;
    }
    z=cgetg(da+3,10); z[1]=evalsigne(1)+evalvarn(varn(x))+evallgef(da+3);
    for (i=0; i<=da; i++) z[i+2]=(long) gmodulcp(stoi(a[i]),p);
    free(a); free(b); 
    lbot=avma; return gerepile(ltop,lbot,gcopy(z));
}

GEN 
gcdmodp(GEN x, GEN y, GEN p){
    GEN a,b,c,*garray[2];
    long ltop=avma,lbot;
    
    if (2*gexpo(p)+6<BITS_IN_LONG) return gcdmodp_withlong(x,y,p);
    a=lift(gmul(x,gmodulcp(gun,p))); b=lift(gmul(y,gmodulcp(gun,p)));
    while (!gcmp0(b)){
        c=resmodp(a,b,p);
        a=b; b=c;
        garray[0]=&a; garray[1]=&b; gerepilemany(ltop,garray,2);
    }
    lbot=avma; return gerepile(ltop,lbot,gmul(a,gmodulcp(gun,p)));
}


GEN
modulargcd(GEN a, GEN b){
    GEN A,B,Cp,p,q,H,g,gp,limit,limit2,ma,mb,*garray[4];
    long m,n,nA,nB,ltop=avma,ltop2,lbot,i;

    if((typ(a)!=10)||(typ(b)!=10)) err(talker,"not a polynomial in modulargcd");
    if(!signe(a)) return gcopy(b);
    if(!signe(b)) return gcopy(a);
    A=gdiv(a,content(a)); B=gdiv(b,content(b)); 
    nA=lgef(A)-3; nB=lgef(B)-3;
    g=ggcd((GEN)A[nA+2],(GEN)B[nB+2]);
    ltop2=avma; q=H=gzero;
    n=1+( (nA<nB) ? nA : nB ); /* min(nA,nB) */
    ma=maxnorm(A); mb=maxnorm(B);
    if (gcmp(ma,mb)==1) limit=mb; else limit=ma;
    limit=gmul2n(gmul(limit,gabs(g,5)),n+1);

    p=gmul2n(gun,BITS_IN_LONG/2-6);
    for (;;){
        p=bigprem(gadd(p,gun)); 
        while (gcmp0(gmod(g,p))) {p=bigprem(gadd(p,gun));}
        gp=gmodulcp(g,p);
        Cp=gcdmodp(A,B,p);
        m=lgef(Cp)-3;
        Cp=gmul(gdiv(gp,(GEN)Cp[m+2]),Cp);
        if (m==0) { avma=ltop; return gun; }
        if (m<n) { q=p; H=Cp; limit=gmul2n(limit,m-n); n=m; }
        else 
            if ((m==n) && (!gcmp0(H))) {
                for (i=0; i<=n; i++)
                    H[2+i]=(long) chinois((GEN)H[2+i],(GEN)Cp[2+i]);
                q=gmul(q,p);
                if (gcmp(limit,q)<=0){
                    limit2=gmul2n(limit,-1);
                    for (i=0; i<=n; i++){
                        p=lift((GEN)H[2+i]);
                        if (gcmp(p,limit2)==1) H[2+i]=lsub(p,q);
                        else H[2+i]=(long) p;
                    }
                    H=gdiv(H,content(H));
                    if ((gcmp0(gmod(A,H))) && (gcmp0(gmod(B,H)))) {
                        lbot=avma;
                        return gerepile(ltop,lbot,gcopy(H)); }
                    else H=gzero; /* failed */
                }
            }
        garray[0]=&H; garray[1]=&p; garray[2]=&q; garray[3]=&limit;
        gerepilemany(ltop2,garray,4);
    }
}
            


/* same as square_free_factorization, but using modulargcd instead of the
gcd of pari */
GEN
mysquare_free_factorization(GEN pol)
/* retourne une matrice a deux colonnes: la 1ere contient les i tels que A_i non
   constant, la deuxieme les A_i, telle que pol=A_i1^i1.A_i2^i2...A_in^in.
   Si pol est constant, retourne la matrice vide. */
{
  long av,tetpil,deg,i,j,va,m;
  GEN p1,p2,x,t1,v1,t,v,*A;

  if(typ(pol)!=10) err(poler7);
  deg=lgef(pol)-3;
  if(deg<1) return cgetg(1,19);
  if(deg==1)
  {
    x=cgetg(3,19);x[1]=lgetg(2,18);x[2]=lgetg(2,18);
    p1=(GEN)x[1];p1[1]=un;p2=(GEN)x[2];p2[1]=lcopy(pol);return x;
  }
  av=avma;va=varn(pol);t1=modulargcd(pol,deriv(pol,va));
  if(isscalar(t1))
  {
    avma=av;
    x=cgetg(3,19);x[1]=lgetg(2,18);x[2]=lgetg(2,18);
    p1=(GEN)x[1];p1[1]=un;p2=(GEN)x[2];p2[1]=lcopy(pol);return x;
  }
  A=(GEN*)newbloc(deg+1);
  v1=gdeuc(pol,t1);v=v1;i=0;
  while(lgef(v)>3)
  {
    v=modulargcd(t1,v1);i++;
    A[i]=gdeuc(v1,v);
    t=gdeuc(t1,v);
    v1=v;t1=t;
  }
  m=0;for(j=1;j<=i;j++) if(isnonscalar(A[j])) m++;
  x=cgetg(3,19);x[1]=lgetg(m+1,18);x[2]=lgetg(m+1,18);m=0;
  for(j=1;j<=i;j++)
  {
    if(isnonscalar(A[j]))
    {
      m++;
      p1=(GEN)x[1];p1[m]=lstoi(j);
      p2=(GEN)x[2];p2[m]=lcopy(A[j]);
    }
  }
  killbloc((GEN)A);
  tetpil=avma;return gerepile(av,tetpil,gcopy(x));
}

/********************************************************************/
/********************************************************************/
/**                                                                **/
/**                 CALCUL DU MODULE DES RACINES                   **/
/**                                                                **/
/********************************************************************/
/********************************************************************/

double 
log2ir(GEN x){
    double l;

    if (signe(x)==0) return (double) -INFINITY;
    if (typ(x)==1){
        if (lgef(x)==3) return (double) log2( (double)(ulong) x[2]);
        l=(double)(ulong) x[2]+
            (double)(ulong) x[3] / exp2((double) BITS_IN_LONG); 
        return log2(l)+ (double) BITS_IN_LONG * (lgef(x)-3.);
    }
    /* else x is real */
    return 1.+ (double) expo(x)+log2( (double)(ulong) x[2]) - (double) BITS_IN_LONG; 
}

double 
mylog2(GEN z){
    double x,y;

    if (typ(z)!=6)
        return log2ir(z);
    /* else z must be complex */
    x=log2ir((GEN) z[1]);
    y=log2ir((GEN) z[2]);
    if (fabs(x-y)>10){
        if (x>y) 
            return x; 
        else
            return y;
    }
    return x+0.5*log2( 1 + exp2(2*(y-x)));
}


long 
findpower(GEN p){
    double x, logbinomial,pente,pentemax=-INFINITY;
    long n=lgef(p)-3,i;

    logbinomial=mylog2((GEN) p[n+2]);
    for (i=n-1; i>=0; i--){
        logbinomial=logbinomial+log2((double) (i+1) / (double) (n-i));
        x=mylog2((GEN) p[2+i])-logbinomial;
        if (x>-INFINITY){
            pente=x/ (double) (n-i);
            if (pente>pentemax) pentemax=pente;
        }       
    }
    return (long) -floor(pentemax);
}

/* returns the exponent for the procedure modulus, from the newton diagram */
long 
polygone_newton(GEN p, long k){
    double *logcoef,pente;
    long n=lgef(p)-3,i,j,h,l,*sommet,pentelong;

    logcoef=(double*) malloc((n+1)*sizeof(double));
    sommet=(long*) malloc((n+1)*sizeof(long)); 
/* sommet[i]=1 si i est un sommet, =0 sinon */
    for (i=0; i<=n; i++){ logcoef[i]=mylog2((GEN)p[2+i]); sommet[i]=0; }
    sommet[0]=1; i=0;
    while (i<n){
        pente=logcoef[i+1]-logcoef[i];
        h=i+1;
        for (j=i+1; j<=n; j++){
            if (pente<(logcoef[j]-logcoef[i])/(double)(j-i)){
                h=j; 
                pente=(logcoef[j]-logcoef[i])/(double)(j-i);
            } 
        }
        i=h;
        sommet[h]=1;
    }
    h=k; while (!sommet[h]) h++;
    l=k-1; while (!sommet[l]) l--;
    pentelong=(long) floor((logcoef[h]-logcoef[l])/(double)(h-l)+0.5);
    free(logcoef); free(sommet);
    return pentelong;
}
    

/* change z into z*2^e, where z is real or complex of real */
void 
myshiftrc(GEN z, long e){

    if (typ(z)==6){
        if (signe((GEN) z[1])!=0) setexpo((GEN)z[1], expo((GEN) z[1])+e);
        if (signe((GEN) z[2])!=0) setexpo((GEN)z[2], expo((GEN) z[2])+e);
    }
    else
        if (signe(z)!=0) setexpo(z,expo(z)+e);
}

/* return z*2^e, where z is integer or complex of integer */
GEN 
myshiftic(GEN z, long e){
    GEN z1;

    if (typ(z)==6){
        z1=cgetg(3,6);
        if (signe((GEN) z[1])!=0)
            z1[1]=(long) mpshift((GEN) z[1],e);
        else 
            z1[1]=lcopy(gzero);
        z1[2]=(long) mpshift((GEN) z[2],e);
        return z1;
    }
    if (signe(z)!=0) 
        return mpshift(z,e);
    else 
        return gcopy(gzero);
}

/* using gexpo in case of complex number is expensive. 
mygexpo is less expensive. */
long 
mygexpo(GEN z){
    long e1,e2;
    
    if (isexactzero(z)) return -INFINITY;
    if (typ(z)==6){
        e1=mygexpo((GEN)z[1]); 
        e2=mygexpo((GEN)z[2]);
        if (e1>e2) return e1; else return e2;
    }
    return gexpo(z);
}

/* return sup(mygexpo(p[i])) for a polynomial p */
long 
exponorm(GEN p){
    long e,e2,i;

    if (gcmp0(p)) return -INFINITY;
    e=mygexpo((GEN)p[2]);
    for (i=3; i<lgef(p); i++){
        e2=mygexpo((GEN)p[i]);
        if (e2>e) e=e2;
    }
    return e;
}


GEN 
mygprecrc(GEN x, long bitprec, long e){
    long tx=typ(x);
    GEN y;

    if (bitprec<=0) bitprec=0; /* should rarely happen */
    switch(tx){
    case 2 :  
        y=cgetr(bitprec/BITS_IN_LONG+3);affrr(x,y);
        if (signe(x)==0) setexpo(y,-bitprec+e);
        break;
    case 6 : y=cgetg(3,6); 
        y[1]=(long) mygprecrc((GEN)x[1],bitprec,e);
        y[2]=(long) mygprecrc((GEN)x[2],bitprec,e);
        break;
    default: y=gcopy(x);
    }
    return y;
}
    

/* gprec behaves badly with the zero for polynomials. 
The second parameter in mygprec is the precision in base 2 */
GEN 
mygprec(GEN x, long bitprec){  
    long  tx=typ(x),lx,i,e;
    GEN   y;
    
    switch(tx){
    case 10: lx=lgef(x);y=cgetg(lx,tx);y[1]=x[1];e=exponorm(x);
        for(i=2;i<lx;i++) y[i]=(long) mygprecrc((GEN)x[i],bitprec,e);
        break;

    default: y=mygprecrc(x,bitprec,0);
    }
    return y;
}


/* the round fonction has a bug in pari. Thus I create mygfloor, using gfloor
which has no bug */
GEN 
mygfloor(GEN z){
    GEN zz;

    if (typ(z)!=6) return gfloor(z);
    zz=cgetg(3,6); zz[1]=lfloor((GEN)z[1]); zz[2]=lfloor((GEN)z[2]);
    return zz;
}

/* returns a polynomial q in (Z[i])[x] keeping bitprec bits of p */
GEN 
eval_rel_pol(GEN p,long bitprec){
    GEN q,r;
    long e=exponorm(p),n=lgef(p)-3,i,ltop=avma,lbot,shift;
    
    q=gprec(p,(long) ((double) bitprec/log2(10.))+2);
    shift=bitprec-e+1;
    r=cgetg(n+3,10); r[1]=p[1];
    for (i=0; i<=n; i++)
        r[2+i]=(long) mygfloor(myshiftic((GEN)q[2+i],shift));
    lbot=avma;
    return gerepile(ltop,lbot,gcopy(r));
}




/* normalize a polynomial p, that is change it with coefficients in Z[i], 
after making product by 2^bitprec */
GEN 
pol_to_gaussint(GEN p, long bitprec){
    long i,n=lgef(p)-3;
    GEN q;

    q=cgetg(n+3,10); q[1]=p[1];
    for (i=0; i<=n; i++){
        myshiftrc((GEN) p[i+2],bitprec); 
        q[i+2]=(long) mygfloor((GEN) p[i+2]);
    }
    return q;
}


/* returns p(R*x)/R^n (in R or R[i]), n=deg(p), bitprec bits of precision */ 
GEN 
homothetie(GEN p, double R, long bitprec){
    GEN q,r,gR,aux;
    long n=lgef(p)-3,i,ltop=avma,lbot;

    gR=mygprec(dbltor(1/R),bitprec);
    q=mygprec(p,bitprec);
    r=cgetg(n+3,10); r[1]=evalsigne(1)+evalvarn(varn(p))+evallgef(n+3);
    aux=gR; r[n+2]=q[n+2];
    for (i=n-1; i>=0; i--){
        r[i+2]=lmul(aux,(GEN)q[i+2]);
        aux=gmul(aux,gR);
    }
    lbot=avma;
    return gerepile(ltop,lbot,gcopy(r));
}

/* change q in 2^(n*e) p(x*2^(-e)), n=deg(q) */
void 
homothetie2n(GEN p, long e){
    long i,n=lgef(p)-3;

    for (i=0; i<n; i++)
        myshiftrc((GEN) p[i+2],(n-i)*e);
}

/* return 2^f * 2^(n*e) p(x*2^(-e)), n=deg(q) */
GEN 
homothetie_gauss(GEN p, long e,long f){
    long i, n=lgef(p)-3;
    GEN q;
    
    q=cgetg(n+3,10);
    q[1]=evalsigne(1)+evalvarn(varn(p))+evallgef(n+3);
    for (i=0; i<=n; i++)
        q[i+2]=(long) myshiftic((GEN) p[i+2],f+(n-i)*e);
    return q;
}

long 
valuation(GEN p){
    long j=0,n=lgef(p)-3;

    while ((j<=n) && isexactzero((GEN)p[j+2])) j++;
    return j;
}

/* provides usually a good lower bound on the largest modulus of the roots,
puts in k an upper bound of the number of roots near the largest root 
at a distance eps */
double 
lower_bound(GEN p, long *k, double eps){
    long n=lgef(p)-3,i,j,ltop=avma;
    GEN a,s,icd;
    double r,*rho;

    if (n<4) { *k=n; return 0. ; }
    a=cgetg(6,10); s=cgetg(6,10);
    rho=(double *) malloc(4*sizeof(double));
    icd=gdiv(gunr,(GEN) p[n+2]);
    for (i=1; i<=4; i++) a[i+1]=lmul(icd,(GEN)p[n+2-i]);
    for (i=1; i<=4; i++){
        s[i+1]=lmulsg(i,(GEN)a[i+1]);
        for (j=1; j<i; j++)
            s[i+1]=ladd((GEN)s[i+1],gmul((GEN)s[j+1],(GEN)a[i+1-j]));
        s[i+1]=lneg((GEN)s[i+1]); 
        r=gtodouble(gabs((GEN) s[i+1],3));
        if (r<=0.)  /* should not be strictly negative */
            rho[i-1]=0. ;
        else
            rho[i-1]=exp(log(r/(double)n)/(double) i);
    }
    r=0.;
    for (i=0; i<4; i++) if (r<rho[i]) r=rho[i];
    if ((r>0.) && (eps<1.2))
        *k=(long) floor((n*rho[0]/r+n)/(1+exp(-eps)*cos(eps))); 
    else
        *k=n;
    free(rho); avma=ltop; return r;
}


/* returns the maximum of the modulus of p with a relative error tau */
GEN
gmax_modulus(GEN p, double tau){
    GEN q,aux,new_gunr;
    long i,j,k,valuat,n=lgef(p)-3,nn,ltop=avma,lbot,bitprec,imax,e;
    double r=0.,rho,tau2,eps=1/log(4./tau);

    tau2=tau/6.;
    bitprec=(long) ((double) n*log2(1./tau2)+3*log2((double) n))+1;
    new_gunr=mygprec(gunr,bitprec+2*n);
    aux=gdiv(new_gunr,(GEN) p[2+n]);
    q=gmul(aux,p); q[2+n]=lcopy(new_gunr);
    k=nn=n;
    e=findpower(q); homothetie2n(q,e); r=-(double) e;
    q=mygprec(q,bitprec+(n<<1));
    q=pol_to_gaussint(q,bitprec);
    imax=(long) ((log(log(4.*n))+log(3./tau))/log(2.))+2;
    for (i=0; i<imax; i++){
        if (i>0) e=findpower(q); else e=0; 
        rho=lower_bound(q,&k,eps);
        if (rho>exp2(-(double) e))
            e=(long) -floor(log2(rho));
        r=r-(double) e/ exp2( (double) i);
        if (i<imax){
            if (k<nn)
                bitprec=(long) ((double) k* log2(1./tau2)+ 
                                (double) (nn-k)*log2(1./eps)+
                                3*log2((double) nn))+1;
            else
                bitprec=(long) ((double) nn* log2(1./tau2)+
                                3.*log2((double) nn))+1;
            q=homothetie_gauss(q,e,bitprec-(long)floor(mylog2((GEN) q[2+nn])+0.5));
            valuat=valuation(q);
            if (valuat>0){
                nn-=valuat; setlgef(q,nn+3);
                for (j=0; j<=nn; j++) q[2+j]=q[2+valuat+j];
            }
            set_karasquare_limit(exponorm(q));
            lbot=avma; q=gerepile(ltop,lbot,graeffe(q));            
            tau2=1.5*tau2; eps=1/log(1./tau2);
        }
    }
    avma=ltop; return(gpui(dbltor(2.),dbltor(r),5));
}

double
max_modulus(GEN p, double tau){
    return gtodouble(gmax_modulus(p,tau));
}

/* return the k-th modulus (in ascending order) of p, rel. error tau*/
double 
modulus(GEN p, long k, double tau){
    GEN q,new_gunr;
    long i,j,kk=k,imax,n=lgef(p)-3,nn,nnn,valuat,ltop=avma,ltop2,lbot,bitprec,decprec,e; 
    double tau2,r;


    tau2=tau/6; nn=n;
    bitprec= (long) ((double) n*(2.+log2(3.*(double) n)+log2(1./tau2)));
    decprec=(long) ((double) bitprec/log2(10.))+1;
    new_gunr=gprec(gunr,decprec);
    ltop2=avma;
    q=gprec(p,decprec);
    q=gmul(new_gunr,q);
    e=polygone_newton(q,k);
    homothetie2n(q,e);
    r=(double) e;
    imax=(long) ((log2(3./tau)+log2(log(4.*(double) n)) ))+1;
    for (i=1; i<imax; i++){
        q=eval_rel_pol(q,bitprec);

        nnn=lgef(q)-3; valuat=valuation(q);
        if (valuat>0){
            kk-=valuat;
            for (j=0; j<=nnn-valuat; j++) q[2+j]=q[2+valuat+j];
            setlgef(q,nnn-valuat+3);
        }
        nn=nnn-valuat;

        set_karasquare_limit(bitprec);
        lbot=avma; q=gerepile(ltop2,lbot,graeffe(q));
        e=polygone_newton(q,kk);
        r=r+e/exp2((double)i);
        q=gmul(new_gunr,q);
        homothetie2n(q,e);

        tau2=1.5*tau2; if (tau2>1.) tau2=1.;
        bitprec= 1+(long) ((double) nn*(2.+log2(3.*(double) nn)+log2(1./tau2)));
    }
    avma=ltop;
    return exp2(-r);
}

/* return the k-th modulus r_k of p, rel. error tau, knowing that 
rmin < r_k < rmax. This helps because the information enable us to use 
less precision... quicker ! */
double 
pre_modulus(GEN p, long k, double tau, double rmin, double rmax){
    GEN q;
    long n=lgef(p)-3,i,imax,imax2,bitprec,ltop=avma,lbot;
    double aux,tau2,R;

    tau2=tau/6.;  aux=sqrt(rmax/rmin)*exp(4*tau2);
    imax=(long) log2(log((double)n)/log(aux)); 
    if (imax<=0) return modulus(p,k,tau);

    R=sqrt(rmin*rmax); 
    bitprec=(long) ((double) n*(2.+log2(aux)+log2(1./tau2)));
    q=homothetie(p,R,bitprec);
    imax2=(long) ((log2(3./tau)+log2(log(4.*(double) n)) ))+1;
    if (imax>imax2) imax=imax2;

    for (i=0; i<imax; i++){
        q=eval_rel_pol(q,bitprec);
        set_karasquare_limit(bitprec);
        lbot=avma; q=gerepile(ltop,lbot,graeffe(q));
        aux=aux*aux*exp(2*tau2);
        tau2=1.5*tau2;
        bitprec= (long) ((double) n*(2.+log2(aux)+log2(1./(1-exp(-tau2)))));
        q=gmul(mygprec(gunr,bitprec),q);
    }

    aux=modulus(q,k,exp2((double)imax)*tau/3.);
    R=R*exp(log(aux)*exp2(-(double)imax));
    avma=ltop;
    return R;
}


/* returns the minimum of the modulus of p with a relative error tau */
double 
min_modulus(GEN p, double tau){
    GEN q;
    long i,n,ltop=avma;
    double r;

    if (isexactzero((GEN)p[2])) return 0.;
    n=lgef(p)-3; q=cgetg(n+3,10);
    q[1]=p[1];
    for (i=0; i<=n; i++)  q[2+i]=p[2+n-i]; /* reciprocal of p */
    r=1./max_modulus(q,tau);
    avma=ltop;
    return r;
}




/* returns k such that r_k e^(-tau) < R < r_{k+1} e^tau.
l is such that you know in advance that l<= k <= n-l */
long 
dual_modulus(GEN p, double R, double tau, long l){
    GEN q;
    long i,j,imax,k,delta_k=0,n=lgef(p)-3,nn,nnn,valuat,ltop=avma,lbot,bitprec,ll=l;
    double logmax,aux,tau2;

    tau2=7.*tau/8.;
    bitprec=6*n-5*l+(long) ((double) n*(log2(1/tau2)+8.*tau2/7.));
    q=homothetie(p,R,bitprec);
    nn=n;
    imax=(long)(log(log(2.*(double)n)/tau2)/log(7./4.)+1);

    for (i=0; i<imax; i++){   
        bitprec=6*nn-5*ll+(long) ((double) nn*(log2(1/tau2)+8.*tau2/7.));

        q=eval_rel_pol(q,bitprec);
        nnn=lgef(q)-3; valuat=valuation(q);
        if (valuat>0){
            delta_k+=valuat;
            for (j=0; j<=nnn-valuat; j++) q[2+j]=q[2+valuat+j];
            setlgef(q,nnn-valuat+3);
        }
        ll= ( (ll-valuat<ll+nnn-n) ? ll-valuat : ll+nnn-n ); 
        /* min(ll-valuat,ll+nnn-n) */
        if (ll<0) ll=0;

        nn=nnn-valuat;
        if (nn==0) return delta_k;


        set_karasquare_limit(bitprec);
        lbot=avma; q=gerepile(ltop,lbot,graeffe(q));    
        tau2=tau2*7./4.;

    }
    k=-1; logmax=- (double) INFINITY;
    for (i=0; i<=lgef(q)-3; i++){
        aux=mylog2((GEN)q[2+i]);
        if (aux>logmax){ logmax=aux; k=i;}
    }
    avma=ltop; return delta_k+k;
}


/********************************************************************/
/********************************************************************/
/**                                                                **/
/**       CALCUL D'UN FACTEUR PAR INTEGRATION SUR LE CERCLE        **/
/**                                                                **/
/********************************************************************/
/********************************************************************/




GEN 
gmulbyi(GEN z){
    GEN aux;
    
    aux=cgetg(3,6);
    if (typ(z)!=6){
        aux[1]=zero;
        aux[2]=(long) z;
    }
    else{
        aux[1]=lneg((GEN)z[2]);
        aux[2]=z[1];
    }
    return aux;
}


void 
fft(GEN Omega, GEN p, GEN f, long step, long l){
    long i,l1,l2,l3,rap=Lmax/l,rapi,step4,ltop,lbot;
    GEN f1,f2,f3,f02,f13,g02,g13,ff;
    
    if (l==2){
        f[0]=ladd((GEN)p[0],(GEN)p[step]);
        f[1]=lsub((GEN)p[0],(GEN)p[step]); return;
    }
    if (l==4){
        f1=gadd((GEN)p[0],(GEN)p[(step<<1)]);
        f3=gadd((GEN)p[step],(GEN)p[3*step]);
        f[0]=ladd(f1,f3);
        f[2]=lsub(f1,f3);
        
        f2=gsub((GEN)p[0],(GEN)p[(step<<1)]);
        f02=gsub((GEN)p[step],(GEN)p[3*step]);
        f02=gmulbyi(f02);
        f[1]=ladd(f2,f02);
        f[3]=lsub(f2,f02);
        return;
    }    
    
    l1=(l>>2); l2=(l1<<1); l3=l1+l2; step4=(step<<2);

    ltop=avma;
    fft(Omega,p,f,step4,l1);
    fft(Omega,p+step,f+l1,step4,l1);
    fft(Omega,p+(step<<1),f+l2,step4,l1);
    fft(Omega,p+3*step,f+l3,step4,l1);

    ff=cgetg(l+1,17);
    for (i=0; i<l1; i++){
        rapi=rap*i;
        f1=gmul((GEN)Omega[rapi],(GEN) f[i+l1]);
        f2=gmul((GEN)Omega[(rapi<<1)],(GEN) f[i+l2]);
        f3=gmul((GEN)Omega[3*rapi],(GEN) f[i+l3]);
        
        f02=gadd((GEN)f[i],f2);
        g02=gsub((GEN)f[i],f2);
        f13=gadd(f1,f3);
        g13=gmulbyi(gsub(f1,f3));

        ff[i+1]=ladd(f02,f13);
        ff[i+l1+1]=ladd(g02,g13);
        ff[i+l2+1]=lsub(f02,f13);
        ff[i+l3+1]=lsub(g02,g13);
    }
    lbot=avma; ff=gerepile(ltop,ltop,gcopy(ff));
    for (i=0; i<l; i++) f[i]=ff[i+1];
}

/* returns a vector RU which contains exp(2*i*k*Pi/NN), k=0..NN-1 */
GEN 
initRU(long NN, long bitprec){
    GEN prim,aux,RU,mygpi;
    long i,N2=(NN>>1),N4=(NN>>2),N8=(NN>>3),decprec;

    RU=cgetg(NN+1,17); RU++;
    mygpi=mppi(bitprec/BITS_IN_LONG+3);    
    aux=gmul(gi,gdivgs(mygpi,NN/2)); /* 2i Pi/NN */
    decprec=(long) ((double) bitprec/log2(10.))+1;
    prim=gexp(aux,decprec);
    RU[0]=lprec(gunr,decprec);

    for (i=1; i<=N8; i++) RU[i]=lmul(prim,(GEN)RU[i-1]);
    for (i=1; i<N8; i++){
        aux=cgetg(3,6);
        aux[1]=((GEN)RU[i])[2]; aux[2]=((GEN)RU[i])[1];
        RU[N4-i]=(long) aux;
    }
    for (i=0; i<N4; i++) RU[i+N4]=(long)gmulbyi((GEN)RU[i]);
    for (i=0; i<N2; i++) RU[i+N2]=lneg((GEN)RU[i]);
    RU--; return RU;
}

/* returns 1 if p has only real coefficients, 0 else */
long 
isreal(GEN p){
    long n=lgef(p)-3,i=0;

    for (i=0;(i<=n) && (typ((GEN)p[i+2])!=6);i++) ;
    if (i<=n) return 0; else return 1;
}


void 
parameters(GEN p, double *mu, double *gamma, 
           long polreal, double param, double param2){ 
    GEN q,pc,Omega,coef,RU,prim,aux,ggamma,gx,mygpi,*garray[2];
    long n=lgef(p)-3,bitprec,NN,K,i,j,ltop2,ltop=avma;
    double lx;

    bitprec=exponorm(p)+(long)param2+8;
    NN=(long) (param*3.14)+1; if (NN<Lmax) NN=Lmax;
    K=NN/Lmax; if (K%2==1) K++; NN=Lmax*K;
    mygpi=mppi(bitprec/BITS_IN_LONG+3);
    aux=gmul(gi,gdivgs(mygpi,NN/2)); /* 2i Pi/NN */
    prim=gexp(aux,(long) ((double) bitprec/log2(10.))+1);
    RU=mygprec(gunr,bitprec);

    Omega=initRU(Lmax,bitprec); Omega++;

    q=mygprec(p,bitprec);
    pc=cgetg(Lmax+1,17); pc++;
    for (i=n+1; i<Lmax; i++) pc[i]=zero;

    coef=cgetg(Lmax+1,17); coef++;
    *mu=(double)INFINITY; *gamma=0.; 
    ggamma=gmul(gzero,gunr);
    if (polreal) K=K/2+1;
    ltop2=avma;
    for (i=0; i<K; i++){

        aux=mygprec(gunr,bitprec);       
        for (j=0; j<=n; j++) {
            pc[j]=lmul((GEN)q[j+2],aux);
            aux=gmul(aux,RU); /* RU = prim^i, aux=prim^(ij) */
        }

        fft(Omega,pc,coef,1,Lmax);    
        for (j=0; j<Lmax; j++){
            aux=gprec((GEN)coef[j],5);
            gx=gabs(aux,5);
            lx=gtodouble(glog(gx,5));
            if (lx<*mu) *mu=lx;
            if ((polreal) && ((i>0) && (i<K-1))){
                gx=gdiv(gdeux,gx); 
                ggamma=gadd(ggamma,gx);
            }
            else{
                ggamma=gadd(ggamma,gdiv(gun,gx));
            }             
        }
        RU=gmul(RU,prim);
        garray[0]=&ggamma; garray[1]=&RU; gerepilemany(ltop2,garray,2);
    }
    ggamma=gdivgs(ggamma,NN);
    *gamma=gtodouble(glog(ggamma,5))/log(2.);
    avma=ltop;
}

/* NN is a multiple of Lmax */
void 
dft(GEN p, long k, long NN, long bitprec, GEN F, GEN H, long polreal){
    GEN Omega,q,qd,pc,pdc,alpha,beta,gamma
        ,RU,aux,*garray[3],U,W,mygpi,prim,prim2;
    long n=lgef(p)-3,i,j,K,ltop,lbot;

    mygpi=mppi(bitprec/BITS_IN_LONG+3);
    aux=gmul(gi,gdivgs(mygpi,NN/2)); /* 2i Pi/NN */
    prim=gexp(aux,(long) ((double) bitprec/log2(10.))+1);
    prim2=mygprec(gunr,bitprec);
    RU=cgetg(n+2,17); RU++;

    Omega=initRU(Lmax,bitprec); Omega++;
    K=NN/Lmax;
    q=mygprec(p,bitprec);
    qd=deriv(q,varn(q));

    pc=cgetg(Lmax+1,17); pc++;
    for (i=n+1; i<Lmax; i++) pc[i]=zero;
    pdc=cgetg(Lmax+1,17); pdc++;
    for (i=n; i<Lmax; i++) pdc[i]=zero;

    alpha=cgetg(Lmax+1,17); alpha++;
    beta=cgetg(Lmax+1,17); beta++;
    gamma=cgetg(Lmax+1,17); gamma++;

    if (polreal) K=K/2+1;

    W=cgetg(k+1,17); U=cgetg(k+1,17);
    for (i=1; i<=k; i++) W[i]=U[i]=zero;

    ltop=avma;
    for (i=0; i<K; i++){ 
        RU[0]=(long) gun;
        for (j=1; j<=n; j++) RU[j]=lmul((GEN)RU[j-1],prim2);
        /* RU[j]=prim^{ij}=prim2^j */

        for (j=0; j<n; j++) pdc[j]=lmul((GEN)qd[j+2],(GEN)RU[j]);
        fft(Omega,pdc,alpha,1,Lmax);
        for (j=0; j<=n; j++) pc[j]=lmul((GEN)q[j+2],(GEN)RU[j]);
        fft(Omega,pc,beta,1,Lmax);
        for (j=0; j<Lmax; j++) gamma[j]=ldiv(gun,(GEN)beta[j]);
        for (j=0; j<Lmax; j++) beta[j]=lmul((GEN)alpha[j],(GEN)gamma[j]);
        fft(Omega,beta,alpha,1,Lmax);
        fft(Omega,gamma,beta,1,Lmax);
        
        if (polreal){ /* case where the polynomial p has real coefficients */
            if ((i>0) && (i<K-1)) {
                for (j=1; j<=k; j++){
                    aux=gmul((GEN)alpha[j+1],(GEN)RU[j+1]);
                    W[j]=ladd((GEN)W[j],gshift(greal(aux),1));
                    aux=gmul((GEN)beta[j],(GEN)RU[j]);
                    U[j]=ladd((GEN)U[j],gshift(greal(aux),1));
                }
            }
            else{
                for (j=1; j<=k; j++){
                    aux=gmul((GEN)alpha[j+1],(GEN)RU[j+1]);
                    W[j]=ladd((GEN)W[j],greal(aux));
                    aux=gmul((GEN)beta[j],(GEN)RU[j]);
                    U[j]=ladd((GEN)U[j],greal(aux));
                }
            }
        }
        else{
            for (j=1; j<=k; j++) {
                W[j]=ladd((GEN)W[j],gmul((GEN)alpha[j+1],(GEN)RU[j+1]));
                U[j]=ladd((GEN)U[j],gmul((GEN)beta[j],(GEN)RU[j]));
            }
        }
        W=gcopy(W); U=gcopy(U); prim2=gmul(prim2,prim);
        garray[0]=&W; garray[1]=&U; garray[2]=&prim2;
        gerepilemany(ltop,garray,3);
    }

    for (i=1; i<=k; i++){
        ltop=avma; aux=(GEN)W[i];
        for (j=1; j<i; j++) aux=gadd(aux,gmul((GEN)W[i-j],(GEN)F[k+2-j]));
        lbot=avma;
        F[k+2-i]=lpile(ltop,lbot,gdivgs(aux,-i*NN));
    }
    for (i=0; i<k; i++){
        ltop=avma; aux=(GEN)U[k-i];
        for (j=1+i; j<k; j++) aux=gadd(aux,gmul((GEN)F[2+j],(GEN)U[j-i]));
        lbot=avma;
        H[i+2]=lpile(ltop,lbot,gdivgs(aux,NN));
    }
}



GEN 
refine_H(GEN F, GEN G, GEN HH, long bitprec, long shiftbitprec){
    GEN H=HH,D,aux,*garray[2];
    long error=0,i=0,bitprec1,bitprec2,lbot,ltop=avma,ltop2;

    D=gsub(gun,gres(gmul(HH,G),F)); error=exponorm(D);
    bitprec2=bitprec+shiftbitprec;

    while (((error>-bitprec) && (i<NEWTON_MAX)) && (error<=0)) {

        garray[0]=&D; garray[1]=&H; gerepilemany(ltop,garray,2);
        bitprec1=-error+shiftbitprec;
        ltop2=avma;
        aux=gmul(mygprec(H,bitprec1),mygprec(D,bitprec1)); 
        aux=mygprec(aux,bitprec1);
        lbot=avma;
        aux=gres(aux,mygprec(F,bitprec1));
        aux=gerepile(ltop2,lbot,aux);

        bitprec1=-error*2+shiftbitprec;
        if (bitprec1>bitprec2) bitprec1=bitprec2;
        H=gadd(mygprec(H,bitprec1),aux);
        D=gsub(gun,gres(gmul(H,G),F)); 
        error=exponorm(D); if (error<-bitprec1) error=-bitprec1;
        i++;
    }
    if (error<=-bitprec/2) { lbot=avma; return gerepile(ltop,lbot,gcopy(H));}
    avma=ltop; return gzero; /* procedure failed */
}

/* return 0 if fails, 1 else */
long 
refine_F(GEN p, GEN *F, GEN *G, GEN H, long bitprec, double gamma){
    GEN pp,FF,GG,r,HH,f0,*garray[4];
    long error,i=0,bitprec1=0,bitprec2,ltop=avma,ltop2,lbot,shiftbitprec,
    shiftbitprec2,n=lgef(p)-3,enh,normF,normG;

    FF=*F; HH=H;
    r=gun; GG=poldivres(p,*F,&r);
    normF=exponorm(FF); normG=exponorm(GG);
    enh=exponorm(H); if (enh<0) enh=0;
    shiftbitprec=normF+2*normG+enh+(long) (4.*log2((double)n)+gamma) +1;
    shiftbitprec2=enh+2*(normF+normG)+(long) (2.*gamma+5.*log2((double)n))+1;
    bitprec2=bitprec+shiftbitprec;
    error=exponorm(r);
    if (error<-bitprec) error=1-bitprec;
    while (((error>-bitprec) && (i<NEWTON_MAX)) && (error<=0)){
        if ((bitprec1==bitprec2) && (i>=2)) {
            shiftbitprec+=n; shiftbitprec2+=n; bitprec2+=n; }
        garray[0]=&FF; garray[1]=&GG; garray[2]=&r; garray[3]=&HH;
        gerepilemany(ltop,garray,4);

        bitprec1=-error+shiftbitprec2;
        HH=refine_H(mygprec(FF,bitprec1),mygprec(GG,bitprec1),
                    mygprec(HH,bitprec1),1-error,shiftbitprec2);
        if (HH==gzero) return 0; /* procedure failed */
        
        bitprec1=-error+shiftbitprec;
        ltop2=avma;
        r=gmul(mygprec(HH,bitprec1),mygprec(r,bitprec1)); 
        r=mygprec(r,bitprec1); 
        lbot=avma;
        f0=gres(r,mygprec(FF,bitprec1)); 
        f0=gerepile(ltop2,lbot,f0);

        bitprec1=-2*error+shiftbitprec;
        if (bitprec1>bitprec2) bitprec1=bitprec2;
        FF=gadd(mygprec(FF,bitprec1),f0);

        bitprec1=-3*error+shiftbitprec;
        if (bitprec1>bitprec2) bitprec1=bitprec2;
        r=gun; pp=mygprec(p,bitprec1);
        GG=poldivres(pp,mygprec(FF,bitprec1),&r);
        error=exponorm(r); 
        if (error<-bitprec1) error=-bitprec1;
        i++;
    }
    if (error<=-bitprec){
        garray[0]=&FF; garray[1]=&GG;
        gerepilemany(ltop,garray,2);
        *F=*(garray[0]); *G=*(garray[1]);
        return 1; /* procedure succeeds */
    }
    avma=ltop; return 0; /* procedure failed */
}

/* returns F and G from the unit circle U such that |p-FG|<2^(-bitprec) |cd|,
where cd is the leading coefficient of p */
void 
split_fromU(GEN p, long k, double delta, long bitprec, 
            GEN *F, GEN *G, double param, double param2){
    GEN pp,FF,GG,H,*garray[2];
    long n=lgef(p)-3,NN,success=0,bitprec2,
    ltop=avma,ltop2,polreal=isreal(p);
    double mu,gamma;

    pp=gdiv(p,(GEN)p[2+n]);
    Lmax=4; while (Lmax<=n) Lmax=(Lmax<<1);
    parameters(pp,&mu,&gamma,polreal,param,param2);

    FF=cgetg(k+3,10); FF[1]=evalsigne(1)+evalvarn(varn(p))+evallgef(k+3);
    FF[k+2]=un;
    H=cgetg(k+2,10); H[1]=evalsigne(1)+evalvarn(varn(p))+evallgef(k+2);

    NN=(long) (0.5/delta); NN+=(NN%2); if (NN<2) NN=2;
    NN=NN*Lmax;
    ltop2=avma;

    while (success!=1){
        avma=ltop2;
        bitprec2=(long) (((double) NN*delta-mu)/log(2.))+exponorm(pp)+8;
        dft(pp,k,NN,bitprec2,FF,H,polreal);
        success=refine_F(pp,&FF,&GG,H,bitprec,gamma);
        NN=(NN<<1);
    }
    GG=gmul(GG,(GEN)p[2+n]);
    garray[0]=&FF; garray[1]=&GG;
    gerepilemany(ltop,garray,2);
    *F=*(garray[0]); *G=*(garray[1]);
}

void 
optimize_split(GEN p, long k, double delta, long bitprec, 
            GEN *F, GEN *G, double param, double param2){
    GEN q,FF,GG;
    long i,n=lgef(p)-3;
    
    if (k<=n/2) split_fromU(p,k,delta,bitprec,F,G,param,param2);
    else{ /* we  start from the reciprocal of p */
        q=cgetg(n+3,10); q[1]=p[1];
        for (i=0; i<=n; i++) q[i+2]=p[n-i+2];
        split_fromU(q,n-k,delta,bitprec,&FF,&GG,param,param2);
        *F=polrecip(GG); *G=polrecip(FF);
    }
}

/********************************************************************/
/********************************************************************/
/**                                                                **/
/**             RECHERCHE DU CERCLE DE SEPARATION                  **/
/**                                                                **/
/********************************************************************/
/********************************************************************/

/* return p(2^e*x) *2^(-n*e) */
GEN 
scalepol2n(GEN p, long e){
    long i,n=lgef(p)-3;
    GEN q;

    q=cgetg(n+3,10); q[1]=p[1];
    for (i=0; i<=lgef(p)-3; i++)
        q[i+2]=lmul2n((GEN)p[i+2],(i-n)*e);
    return q;
}

/* returns p(x/R)*R^n */
GEN 
scalepol(GEN p, GEN R, long bitprec){
    GEN q,r,aux,gR;
    long n=lgef(p)-3,i,ltop=avma,lbot;

    gR=mygprec(R,bitprec);
    q=mygprec(p,bitprec);
    r=cgetg(n+3,10); r[1]=evalsigne(1)+evalvarn(varn(p))+evallgef(n+3);
    aux=gR;
    r[n+2]=q[n+2];
    for (i=n-1; i>=0; i--){
        r[i+2]=lmul(aux,(GEN)q[i+2]);
        aux=gmul(aux,gR);
    }
    lbot=avma;
    return gerepile(ltop,lbot,gcopy(r));
}

        
/* returns q(x)=p(b), where b is (usually) a polynomial  */
GEN 
shiftpol(GEN p, GEN b){
    long i,n=lgef(p)-3,ltop=avma,lbot;
    GEN q,r;
    
    q=gzero;
    for (i=n; i>=0; i--){
        r=gmul(q,b);
        lbot=avma;
        q=gadd((GEN)p[i+2],r);
        if (i%2==1) q=gerepile(ltop,lbot,q);
    }
    return gerepile(ltop,lbot,q);
}

/* return (aX-1)^n * p[ (X-a) / (aX-1) ] */
GEN 
conformal_pol(GEN p, GEN a, long bitprec){
    GEN r,pui,num,aux,*garray[2];
    long n=lgef(p)-3,ltop=avma,lbot,ltop2,i;

    pui=cgetg(4,10); pui[1]=evalsigne(1)+evalvarn(varn(p))+evallgef(4);
    pui[2]=(long) mygprec(gneg(gunr),bitprec); pui[3]=lconj(a);
    aux=pui;
    num=cgetg(4,10); num[1]=pui[1];
    num[2]=lneg(a); num[3]=(long) mygprec(gunr,bitprec);
    ltop2=avma;
    r=(GEN)p[2+n];
    for (i=n-1; i>=0; i--){
        r=gadd(gmul(r,num),gmul(aux,(GEN) p[2+i]));
        aux=gmul(pui,aux);
        if (i%2==1)
            {garray[0]=&r; garray[1]=&aux; gerepilemany(ltop2,garray,2);}
    }
    lbot=avma;
    return gerepile(ltop,lbot,gcopy(r));
}


/* apply the conformal mapping then split from U */
void conformal_mapping(GEN p, long k, GEN u, long bitprec, double aux,GEN *F, GEN *G){
    long bitprec2,bitprec3,n=lgef(p)-3,decprec,i,ltop1=avma,ltop,lbot;
    GEN q,FF,GG,a,R, *garray[2];
    double rmin,rmax,rho,delta,aux2,param,param2,r1,r2;

    bitprec2=bitprec+(long) (n*(2.*log2(2.732)+log2(1.5)))+1;
    a=gsqrt(stoi(3),10);
    a=gmul(mygprec(a,bitprec2),mygprec(u,bitprec2));
    a=gdivgs(a,-6);  /* a=-u/2/sqrt(3) */
    
    ltop=avma;
    q=mygprec(p,bitprec2);
    lbot=avma;
    q=gerepile(ltop,lbot,conformal_pol(q,a,bitprec2));
    for (i=1; i<=n; i++) { /* updating array radius */
        if (radius[i]!=0.){
            aux2=radius[i]*radius[i];
            aux2=2.*(aux2-1)/(aux2-3.*radius[i]+3.);
            radius[i]=sqrt(1+aux2);
        }
    }

    if (k>1){
        i=k-1; while ((i>0) && (radius[i]==0.)) i--;
        r1=radius[i]; r2=radius[k];
        rmin=pre_modulus(q,k,aux/10.,r1,r2); 
    }
    else
        rmin=min_modulus(q,aux/10.);
    radius[k]=rmin; 

    if (k+1<n){
        i=k+2; while ((i<=n) && (radius[i]==0.)) i++;
        r2=radius[i]; r1=radius[k+1];
        rmax=pre_modulus(q,k+1,aux/10.,r1,r2); 
    }
    else /* k+1=n */
        rmax=max_modulus(q,aux/10.);
    radius[k+1]=rmax; 

    rho=sqrt(rmin*rmax); delta=0.5*log(rmax/rmin);
    if (delta>1.) delta=1.;

    if (rho<1.) bitprec3=(long) ((double)n*log2(1./rho))+1;
    else bitprec3=(long) ((double)n*log2(rho))+1;
    bitprec2=bitprec2+bitprec3;

    R=mygprec(dbltor(1/rho),bitprec2);
    q=scalepol(q,R,bitprec2);
    garray[0]=&R; garray[1]=&q; gerepilemany(ltop,garray,2);


    aux2=radius[k];
    for (i=k-1; i>=1; i--){
        if (radius[i]==0.) radius[i]=aux2;
        else aux2=radius[i];
    }
    aux2=radius[k+1];
    for (i=k+1; i<=n; i++){
        if (radius[i]==0.) radius[i]=aux2;
        else aux2=radius[i];
    }
    param=0.; param2=0.;
    for (i=1; i<=n; i++) {
        radius[i]=radius[i]/rho;
        aux2=fabs(1-radius[i]);
        param+=1./aux2;
        if (aux2<1.) param2-=log2(aux2);
    }
    optimize_split(q,k,delta,bitprec2,&FF,&GG,param,param2);
    bitprec2=bitprec2+n;
    R=gdiv(gun,R);
    FF=scalepol(FF,R,bitprec2); GG=scalepol(GG,R,bitprec2);
    garray[0]=&FF; garray[1]=&GG; gerepilemany(ltop,garray,2);

    a=mygprec(a,bitprec2);    
    FF=conformal_pol(FF,a,bitprec2); GG=conformal_pol(GG,a,bitprec2);
    a=gdiv(gun,gsub(gun,gmul(a,gconj(a)))); 
    a=glog(a,(long) (bitprec2/log2(10.))+1);

    decprec=(long) ((bitprec+n)/log2(10.))+1;
    FF=gmul(FF,gexp(gmulgs(a,k),decprec)); 
    GG=gmul(GG,gexp(gmulgs(a,n-k),decprec)); 


    FF=mygprec(FF,bitprec+n); GG=mygprec(GG,bitprec+n);
    garray[0]=&FF; garray[1]=&GG; gerepilemany(ltop1,garray,2);
    *F=*garray[0]; *G=*garray[1];
}
    



/* split p, this time with no scaling . returns in F and G two polynomials 
such that |p-FG|< 2^(-bitprec)|p| */
void split_2(GEN p, long bitprec, double thickness, GEN *F, GEN *G){
    double rmin,rmax,rho,kappa,aux,delta,param,param2,r1,r2;
    long n=lgef(p)-3,i,j,k,disti,diste,bitprec2,ltop1=avma,ltop;
    GEN q,FF,GG,*garray[2],R;


    radius=(double *) malloc((n+1)*sizeof(double));
    for (i=2; i<n; i++) radius[i]=0.;
    rmin=min_modulus(p,thickness/(double) n/4.);
    rmax=max_modulus(p,thickness/(double) n/4.);
    i=1; j=n; radius[1]=rmin; radius[n]=rmax;
    rho=sqrt(rmin*rmax);
    k=dual_modulus(p,rho,thickness/(double) n/4.,1);
    if ((k<n/5) || ((n/2<k) && (k<(4*n)/5))) { rmax=rho; j=k+1;radius[j]=rho;}
    else { rmin=rho; i=k; radius[i]=rho; }
    while (j>i+1){
        disti= ( (i<n-j) ? i : n-j ); /* min(i,n-j) */
        diste= ( (j<n-i) ? j : n-i );
        kappa=1.-log(1.+(double)disti)/log(1.+(double)diste);
        if (i+j<n+1) 
            rho=exp( (log(rmin)+log(rmax)*(1+kappa))/(2+kappa));
        else
            rho=exp( (log(rmin)*(1+kappa)+log(rmax))/(2+kappa));
        if (i+j==n+1) rho=sqrt(rmin*rmax);
        aux=log(rmax/rmin)/(j-i)/4.;
        k=( (i<n+1-j) ? i : n+1-j); /* min(i,n+1-j) */
        k=dual_modulus(p,rho,aux,k);
        if ((k-i)<(j-k-1)) {rmax=rho; j=k+1; radius[j]=rho*exp(-aux); }
        else {
            if ((k-i)>(j-k-1)) {rmin=rho; i=k; radius[i]=rho*exp(aux);}
            else{
                if (2*k>n) {rmax=rho; j=k+1; radius[j]=rho*exp(-aux);}
                else {rmin=rho; i=k; radius[i]=rho*exp(aux); }
            }
        }
    }
    
    aux=log(rmax/rmin);

    if (step4==0){
        if (k>1){
            i=k-1; while ((i>0) && (radius[i]==0.)) i--;
            r1=radius[i]; r2=radius[k];
            rmin=pre_modulus(p,k,aux/10.,r1,r2); 
        }
        else /* k=1 */
            rmin=min_modulus(p,aux/10.);
        radius[k]=rmin; 
        
        if (k+1<n){
            i=k+2; while ((i<=n) && (radius[i]==0.)) i++;
            r2=radius[i]; r1=radius[k+1];
            rmax=pre_modulus(p,k+1,aux/10.,r1,r2); 
        }
        else /* k+1=n */
            rmax=max_modulus(p,aux/10.);
        radius[k+1]=rmax; 
        rho=sqrt(rmin*rmax); delta=0.5*log(rmax/rmin);
        

        aux=radius[k];
        for (i=k-1; i>=1; i--){
            if (radius[i]==0.) radius[i]=aux;
            else aux=radius[i];
        }
        aux=radius[k+1];
        for (i=k+1; i<=n; i++){
            if (radius[i]==0.) radius[i]=aux;
            else aux=radius[i];
        }
        param=0.; param2=0.;
        for (i=1; i<=n; i++) {
            radius[i]=radius[i]/rho;
            aux=fabs(1.-radius[i]);
            param+=1./aux;
            if (aux<1) param2-=log2(aux);
        }
        if (rho<1.) bitprec2=(long) ((double)n*log2(1./rho))+1;
        else bitprec2=(long) ((double)n*log2(rho))+1;
        bitprec2=bitprec+bitprec2;
        
        R=mygprec(dbltor(1./rho),bitprec2);
        q=scalepol(p,R,bitprec2);
        optimize_split(q,k,delta,bitprec2,&FF,&GG,param,param2);
        ltop=avma; bitprec2+=n;
        R=gdiv(gun,R); 
        FF=scalepol(FF,R,bitprec2); GG=scalepol(GG,R,bitprec2);
        FF=mygprec(FF,bitprec+n); GG=mygprec(GG,bitprec+n);
        
    }
    else{
        rho=sqrt(rmax*rmin);
        if (rho<1.) bitprec2=(long) ((double)n*log2(1./rho))+1;
        else bitprec2=(long) ((double)n*log2(rho))+1;
        bitprec2=bitprec+bitprec2;
        
        R=mygprec(dbltor(1/rho),bitprec2);
        q=scalepol(p,R,bitprec2);
        for (i=1; i<=n; i++)
            if (radius[i]!=0.) radius[i]=radius[i]/rho;

        conformal_mapping(q, k, u, bitprec2, aux, &FF, &GG);
        bitprec2=bitprec2+n;
        R=gdiv(gun,mygprec(R,bitprec2)); 
        FF=scalepol(FF,R,bitprec2); GG=scalepol(GG,R,bitprec2);
        FF=mygprec(FF,bitprec+n); GG=mygprec(GG,bitprec+n);
    }
    
    garray[0]=&FF; garray[1]=&GG;
    gerepilemany(ltop1,garray,2);
    *F=*(garray[0]); *G=*(garray[1]);
    free(radius);
}

/* procedure corresponding to steps 5,6,.. page 44 in the RR n. 1852 */
/* put in F and G two polynomial such that |p-FG|<2^(-bitprec)|p|
where the maximum modulus of the roots of p is <=1 and the sum of roots
is zero */

void split_1(GEN p, long bitprec, GEN *F, GEN *G){
    long bitprec2,bitprec3,n=lgef(p)-3,i
        ,ltop1=avma,ltop,imax,polreal=isreal(p);
    GEN q,qq,newq,FF,GG,*garray[2],v,gr,r;
    double rmax,rmin,thickness;

    r=gmax_modulus(p,0.01); 
    bitprec2=bitprec+n;
    gr=mygprec(gdiv(gun,r),bitprec2);
    q=scalepol(p,gr,bitprec2);

    bitprec2=bitprec+exponorm(q)-exponorm(p);

    bitprec3=bitprec2+(long)((double)n*2.*log2(3.)+1);
    v=cgetg(5,17); v++;
    v[0]=lmulgs(mygprec(gunr,bitprec3),2); v[1]=lneg((GEN)v[0]);
    v[2]=lmul((GEN)v[0],gi); v[3]=lneg((GEN)v[2]);
    if (polreal) imax=3; else imax=4;
    q=mygprec(q,bitprec3);
    thickness=1.;
    for (i=0; i<imax; i++){
        qq=shiftpol(q,gadd(polx[varn(p)],(GEN)v[i]));
        rmin=min_modulus(qq,0.05); 
        if ((3./rmin)>thickness){
            rmax=max_modulus(qq,0.05);
            if (rmax/rmin>thickness){
                thickness=rmax/rmin; 
                newq=qq;
                u=(GEN)v[i];
            }
        }
        if (thickness>2.) i=imax;
        if ((polreal) && ((i==1) && (thickness>1.5))) i=imax;
    }
    bitprec3=bitprec2+(long)((double) n*log2(3.)+1)+exponorm(newq)-exponorm(q);
    split_2(newq,bitprec3,log(thickness),&FF,&GG);
    ltop=avma;
    u=gsub(polx[varn(p)],mygprec(u,bitprec3));
    FF=shiftpol(FF,u); GG=shiftpol(GG,u);
    garray[0]=&FF; garray[1]=&GG; gerepilemany(ltop,garray,2);
    ltop=avma;
    bitprec2=bitprec2+exponorm(FF)+exponorm(GG)-exponorm(q);
    gr=gdiv(gun,gr);
    FF=scalepol(FF,gr,bitprec2); GG=scalepol(GG,gr,bitprec2);
    
    
    garray[0]=&FF; garray[1]=&GG;
    gerepilemany(ltop1,garray,2);
    *F=*(garray[0]); *G=*(garray[1]);   
}

/* put in F and G two polynomials such that |P-FG|<2^(-bitprec)|P|,
where the maximum modulus of the roots of p is <0.5 */
void split_0_2(GEN p, long bitprec, GEN *F, GEN *G){
    GEN q,b,FF,GG;
    long n=lgef(p)-3,k,bitprec2,i;
    double auxnorm;

    step4=1;
    auxnorm=(double) n*log2(1+exp2(mylog2((GEN)p[n+1])
                                   -mylog2((GEN)p[n+2]))/(double)n);
    bitprec2=bitprec+1+(long) (log2((double)n)+auxnorm);
                
    q=mygprec(p,bitprec2);
    b=gdivgs(gdiv((GEN)q[n+1],(GEN)q[n+2]),-n);
    q=shiftpol(q,gadd(polx[varn(p)],b));
 
    k=0; while 
        (mygexpo((GEN)q[k+2])<-(bitprec2+2*(n-k)+exponorm(q))
         || gcmp0((GEN)q[k+2])) k++;
    if (k>0){
        if (k>n/2) k=n/2;
        bitprec2+=(k<<1);
        FF=cgetg(k+3,10); FF[1]=evalsigne(1)+evalvarn(varn(p))+evallgef(k+3);
        for (i=0; i<k; i++) FF[i+2]=zero;
        FF[k+2]=(long) mygprec(gunr,bitprec2);
        GG=cgetg(n-k+3,10);GG[1]=evalsigne(1)+evalvarn(varn(p))+evallgef(n-k+3);
        for (i=0; i<=n-k; i++) GG[i+2]=q[i+k+2];
        b=gsub(polx[varn(p)],mygprec(b,bitprec2));
        GG=mygprec(GG,bitprec2);
        FF=shiftpol(FF,b); GG=shiftpol(GG,b);
    }
    else{
        split_1(q,bitprec2,&FF,&GG);
        bitprec2=bitprec+exponorm(FF)+exponorm(GG)-exponorm(p)+(long)auxnorm+1; 
        b=gsub(polx[varn(p)],mygprec(b,bitprec2)); 
        FF=mygprec(FF,bitprec2); GG=mygprec(GG,bitprec2);
        FF=shiftpol(FF,b); GG=shiftpol(GG,b);
    }
    *F=FF; *G=GG;
}


/* put in F and G two polynomials such that |P-FG|<2^(-bitprec)|P|,
where the maximum modulus of the roots of p is <2 */
void split_0_1(GEN p, long bitprec, GEN *F, GEN *G){
    GEN q,FF,GG,*garray[2];
    long n=lgef(p)-3,bitprec2,ltop=avma,normp,pow;
    double aux;

    normp=exponorm(p);
    aux=exp2(mylog2((GEN)p[n+1])-mylog2((GEN)p[n+2]))/(double)n;
    pow=2;
    if (aux<=2.5){ split_0_2(p,bitprec,F,G); return; }
    if (aux<=1.) pow=1; 
    q=scalepol2n(p,pow); /* q(x)=4^(-n) p(4*x) */
    bitprec2=bitprec+pow*n+exponorm(q)-normp;
    q=mygprec(q,bitprec2);
    split_1(q,bitprec2,&FF,&GG);
    FF=scalepol2n(FF,-pow); GG=scalepol2n(GG,-pow);
    bitprec2=bitprec+exponorm(FF)+exponorm(GG)-normp;
    FF=mygprec(FF,bitprec2); GG=mygprec(GG,bitprec2);
    garray[0]=&FF; garray[1]=&GG;
    gerepilemany(ltop,garray,2);
    *F=*(garray[0]); *G=*(garray[1]);
}

/* put in F and G two polynomials such that |P-FG|<2^(-bitprec)|P| */
void split_0(GEN p, long bitprec, GEN *F, GEN *G){
    GEN FF,GG,q, *garray[2];
    long n=lgef(p)-3,k=0,i,ltop=avma;
    double R1,R2;

    while (mygexpo((GEN)p[k+2])<-bitprec) k++;
    if (k>0){
        if (k>n/2) k=n/2;
        FF=cgetg(k+3,10); FF[1]=evalsigne(1)+evalvarn(varn(p))+evallgef(k+3);
        for (i=0; i<k; i++) FF[i+2]=lcopy(gzero);
        FF[k+2]=(long) mygprec(gunr,bitprec);
        GG=cgetg(n-k+3,10);GG[1]=evalsigne(1)+evalvarn(varn(p))+evallgef(n-k+3);
        for (i=0; i<=n-k; i++) GG[i+2]=p[i+k+2];
    }
    else{
        R1=max_modulus(p,0.05);
        if (R1<1.9){ split_0_1(p,bitprec,&FF,&GG); }
        else{
            q=cgetg(n+3,10); q[1]=p[1];
            for (i=0; i<=n; i++) q[i+2]=p[n-i+2]; /* p^* with copy of ptr */
            R2=max_modulus(q,0.05);
            if (R2<1.9){
                split_0_1(q,bitprec,&FF,&GG);
                FF=polrecip(FF); GG=polrecip(GG);
            }
            else{
                step4=0;
                split_2(p,bitprec,1.2837,&FF,&GG);
            }
        }
    }
    garray[0]=&FF;
    garray[1]=&GG;

    gerepilemany(ltop,garray,2);
    *F=*(garray[0]);
    *G=*(garray[1]);
}

/********************************************************************/
/********************************************************************/
/**                                                                **/
/**     CALCUL A POSTERIORI DE L'ERREUR ABSOLUE SUR LES RACINES    **/
/**                                                                **/
/********************************************************************/
/********************************************************************/

GEN 
root_error(long n, long k, GEN roots_pol, GEN sigma, GEN shatzle){
    GEN rho,d,eps,epsbis,eps2,prod,rap,aux;
    long i,j,m,ltop=avma,lbot;

    d=cgetg(n+1,17);
    for (i=1; i<=n; i++){
        if (i!=k) { 
            aux=gsub((GEN)roots_pol[i],(GEN)roots_pol[k]);
            d[i]=(long) gabs(mygprec(aux,31),4);
        }
    }
    rho=gabs(mygprec((GEN)roots_pol[k],31),4); 
    if (gcmp(rho,dbltor(1.))==-1) rho=gun;
    eps=gmul(rho,shatzle);
    aux=gmul(gpui(rho,stoi(n),4),sigma);

    for (j=1; (j<=2) || ((gcmp(rap,dbltor(1.2))==1) && (j<=5)); j++){
        m=n; prod=gun;
        epsbis=gdivgs(gmulgs(eps,5),4);
        for (i=1; i<=n; i++){
            if ((i!=k) && (gcmp((GEN)d[i],epsbis)==1)){
                m--;
                prod=gmul(prod,gsub((GEN)d[i],eps));
            }
        }
        eps2=gdiv(gmul2n(aux,2*m-2),prod);
        eps2=gpui(eps2,dbltor(1./m),4);
        rap=gdiv(eps,eps2);
        eps=eps2;
    }
    lbot=avma;
    return gerepile(ltop,lbot,gcopy(eps));
}

/* round a complex or real number x to an absolute value of 2^(-e) */
GEN 
mygprec_absolute(GEN x, long bitprec){
    long  tx=typ(x),e;
    GEN   y;
    
    switch(tx){
    case 2: e=mygexpo(x);
      if ((e<-bitprec)||(signe(x)==0)) {
	y=dbltor(0.);setexpo(y,-bitprec);
      }
        else{
            y=mygprec(x,bitprec+e);
        }
        break;
    case 6: if (mygexpo((GEN)x[2])<-bitprec) 
        y=mygprec_absolute((GEN)x[1],bitprec); 
        else{
            y=cgetg(3,6); 
            y[1]=(long) mygprec_absolute((GEN)x[1],bitprec);                
            y[2]=(long) mygprec_absolute((GEN)x[2],bitprec);
        }
        break;
        
    default: y=mygprec(x,bitprec);
    }
    return y;
}

long
a_posteriori_errors(GEN p, GEN roots_pol, long err){
    GEN sigma,shatzle,x;
    long ltop,i,n=lgef(p)-3,e,e_max;

    ltop=avma;
    sigma=mygprec(dbltor(1.),10);
    setexpo(sigma,err+(long) log2((double) n)+1);
    shatzle=gdiv(gpui(sigma,dbltor(1./n),4),
                 gsub(gpui(gsub(gun,sigma),dbltor(1./n),4),
                      gpui(sigma,dbltor(1./n),4)));
    shatzle=gmul2n(shatzle,1);
    if (gcmp0(shatzle)) {
        output(shatzle);
        printf("expo:%ld\n",expo(shatzle));
    }
    e_max=-INFINITY;
    for (i=1; i<=n; i++){
        x=root_error(n,i,roots_pol,sigma,shatzle);
        e=mygexpo(x); if (e>e_max) e_max=e;
        roots_pol[i]=(long) mygprec_absolute((GEN)roots_pol[i],-e);
    }
    return e_max;
}

/********************************************************************/
/********************************************************************/
/**                                                                **/
/**                           MAIN                                 **/
/**                                                                **/
/********************************************************************/
/********************************************************************/


/* compute roots in roots_pol so that |P-L_1...L_n|<2^(-bitprec)|P| , and
returns prod (x-roots_pol[i]) for i=1..degree(p) */ 
GEN
split_complete(GEN p, long bitprec, GEN *roots_pol, long *iroots){
    long n=lgef(p)-3,decprec,ltop;
    GEN F,G,a,b,*garray[3],m1,m2,m;

    if (n==1){
        a=gneg(gdiv((GEN)p[2],(GEN)p[3]));
        (*iroots)++;
        (*roots_pol)[*iroots]=(long) a;
        return p;
    }
    if (n==2){
        ltop=avma;
        F=gsub(gsqr((GEN)p[3]),gmul2n(gmul((GEN)p[2],(GEN)p[4]),2));
        decprec=(long) ((double) bitprec/log2(10.))+1;
        F=gsqrt(F,decprec);
        a=gneg(gdiv(gadd((GEN)p[3],F),gmul2n((GEN)p[4],1)));
        b=gdiv(gsub(F,(GEN)p[3]),gmul2n((GEN)p[4],1));
        
        garray[0]=&a;
        garray[1]=&b;
        gerepilemany(ltop,garray,2);
        (*iroots)++; (*roots_pol)[*iroots]=(long) a;
        (*iroots)++; (*roots_pol)[*iroots]=(long) b;
        m=gmul(gsub(polx[varn(p)],mygprec(a,3*bitprec)),
               gsub(polx[varn(p)],mygprec(b,3*bitprec)));
        return gmul(m,(GEN)p[4]);
    }
    ltop=avma;
    split_0(p,bitprec,&F,&G);
    m1=split_complete(F,bitprec,roots_pol,iroots);
    m2=split_complete(G,bitprec,roots_pol,iroots);
    m=gmul(m1,m2);
    *roots_pol=gcopy(*roots_pol);
    garray[0]=&(*roots_pol); garray[1]=&m;
    gerepilemany(ltop,garray,2);
    *roots_pol=*(garray[0]);
    return m;
}


/* compute a bound on the maximum modulus of roots of p */
GEN 
cauchy_bound(GEN p){
    long i,n=lgef(p)-3;
    GEN x=gzero,y,lc;

    lc=gabs((GEN)p[n+2],5); /* leading coefficient */
    lc=gdiv(dbltor(1.),lc); 
    for (i=0; i<n; i++){
        y=gmul(gabs((GEN) p[i+2],5),lc);
        y=gpui(y,dbltor(1./(n-i)),5);
        if (gcmp(y,x)==1) x=y;
    }
    return x;
}

GEN 
mygprecrc_special(GEN x, long bitprec, long e){
    long tx=typ(x),lx,ex;
    GEN y;

    if (bitprec<=0) bitprec=0; /* should not happen */
    switch(tx){
    case 2 :  
        lx=bitprec/BITS_IN_LONG+3;
        if (lx<lg(x)) lx=lg(x);
        y=cgetr(lx);affrr(x,y);
        ex=-bitprec+e;
        if ((signe(x)==0) && (gexpo(x)>ex)) setexpo(y,ex);
        break;
    case 6 : y=cgetg(3,6); 
        y[1]=(long) mygprecrc_special((GEN)x[1],bitprec,e);
        y[2]=(long) mygprecrc_special((GEN)x[2],bitprec,e);
        break;
    default: y=gcopy(x);
    }
    return y;
}
    

/* like mygprec but keep at least the same precision as before */
GEN 
mygprec_special(GEN x, long bitprec){  
    long  tx=typ(x),lx,i,e;
    GEN   y;
    
    switch(tx){
    case 10: lx=lgef(x);y=cgetg(lx,tx);y[1]=x[1];e=exponorm(x);
        for(i=2;i<lx;i++) y[i]=(long) mygprecrc_special((GEN)x[i],bitprec,e);
        break;

    default: y=mygprecrc_special(x,bitprec,0);
    }
    return y;
}


GEN
all_roots(GEN p, long bitprec){
    GEN q,roots_pol,m;
    long bitprec2,n=lgef(p)-3,iroots,i,j,e,ltop=avma,lbot;

    for (j=1; j<=10; j++){
      avma=ltop; iroots=0;
        roots_pol=cgetg(n+1,17); for (i=1; i<=n; i++) roots_pol[i]=zero;
        
        e=2*mygexpo(cauchy_bound(p)); if (e<0) e=0;
        bitprec2=bitprec+(1<<j)*n+exponorm(p)-mygexpo((GEN)p[2+n])
            +(long) log2(n)+1+e;
        q=gmul(mygprec(gunr,bitprec2),mygprec(p,bitprec2));
        m=split_complete(q,bitprec2,&roots_pol,&iroots);
        e=exponorm(gsub(mygprec_special(p,bitprec2),m))
            -mygexpo((GEN)q[2+n])+(long) log2((double)n)+1;
        if (e<-2*bitprec2) e=-2*bitprec2; /* to avoid e=-INFINITY */
        e=a_posteriori_errors(q,roots_pol,e);
      if (e<-bitprec){
            lbot=avma; return gerepile(ltop,lbot,gcopy(roots_pol));
        }
    }
    err(talker,"bug in all_roots");return gnil;
}


/* true if x is an exact scalar, that is integer or rational */
int 
isexactscalar(GEN x){
    long tx=typ(x);

    if ((tx==1) || ((tx==4) || (tx==5))) return 1;
    else return 0;
}

int 
isexactpol(GEN p){
    long i,n=lgef(p)-3;

    for (i=0; i<=n; i++){
        if (isexactscalar((GEN)p[i+2])==0) return 0;
    }
    return 1;
}

long 
isvalidcoeff(GEN x){
    long tx=typ(x);
    
    switch(tx){
    case 1: 
    case 2: 
    case 4:
    case 5: return 1;
    case 6: if ((isvalidcoeff((GEN)x[1])) && (isvalidcoeff((GEN)x[2])))
        return 1;
        else return 0;
    default: return 0;
    }
}

long 
isvalidpol(GEN p){
    long i,n;

    n=lgef(p)-3;
    for (i=0; i<=n; i++) if (!isvalidcoeff((GEN)p[2+i])) return 0;
    return 1;
}


GEN
solve_exact_pol(GEN p, long bitprec){
    GEN S,multiplicities,factors,roots_pol,roots_fact;
    long i,j,k,m,n,iroots,ltop=avma,lbot;

    n=lgef(p)-3;

    iroots=0;
    roots_pol=cgetg(n+1,17); for (i=1; i<=n; i++) roots_pol[i]=zero;

    S=mysquare_free_factorization(p);
    multiplicities=(GEN) S[1];
    factors=(GEN) S[2];
    for (i=1; i<=lg(factors)-1; i++){
        roots_fact=all_roots((GEN)factors[i],bitprec);
        n=lgef((GEN)factors[i])-3; m=gtolong((GEN)multiplicities[i]);
        for (j=1; j<=n; j++){
            for (k=1; k<=m; k++){ 
                iroots++; roots_pol[iroots]=roots_fact[j];
            }
        }
    }
    lbot=avma; return gerepile(ltop,lbot,gcopy(roots_pol));
}

/* return the roots of p with absolute error bitprec */
GEN 
roots_com(GEN p, long l){
    long bitprec;

    if (typ(p)!=10) {
        if (isvalidcoeff(p)){
            if (!gcmp0(p)) return cgetg(1,17); /* constant polynomial */
            err(talker,"zero polynomial in roots");
        }
        err(talker,"not polynomial argument in roots");
    }
    if (!isvalidpol(p)) err(talker,"invalid coefficients in roots");
    if (lgef(p)==3) {
        if (!gcmp0((GEN)p[2])) return cgetg(1,17); /* constant polynomial */
        err(talker,"zero polynomial in roots");
    }
    if (lgef(p)==2) err(talker,"zero polynomial in roots");
    if (l<1) l=1;
    bitprec=l*BITS_IN_LONG;
    gunr=cgetr(l);affsr(1,gunr);
    gzeror=cgetr(l);affsr(0,gzeror);
    if (isexactpol(p)) return solve_exact_pol(p,bitprec);
    else return all_roots(p,bitprec); 
}

GEN
tocomplex(GEN x, long l)
{
  GEN y;

  y=cgetg(3,6);y[1]=lgetr(l);y[2]=lgetr(l);
  switch(typ(x))
    {
      case 1: case 4: case 5: gaffect(x,(GEN)y[1]);affsr(0,(GEN)y[2]);break;
      case 2: affrr(x,(GEN)y[1]);affsr(0,(GEN)y[2]);break;
      case 6: gaffect((GEN)x[1],(GEN)y[1]);gaffect((GEN)x[2],(GEN)y[2]);break;
      default: err(talker,"incorrect type in iscomplex");
    }
  return y;
}

int
isrealappr(GEN x, long l)
{
  int fl;
  long tx,lx,i;
  switch(tx=typ(x))
    {
      case 1: case 2: case 4: case 5: return 1;
      case 6: return (gcmp0((GEN)x[2])||(gexpo((GEN)x[2])<(2-((l-2)<<TWOPOTBITS_IN_LONG))));
      case 8: err(impl,"isrealappr for type 8");
      case 10: case 11: case 13: case 14: case 17: case 18: case 19:
	if(tx==10) lx=lgef(x);
	if((tx==11)&&(!signe(x))) lx=2;
	fl=1;for(i=lontyp[tx];(i<lx)&&fl;i++) fl=isrealappr((GEN)x[i],l);
	return fl;
      default: err(talker,"incorrect type in isrealappr");return 0;
    }
}

int
isconj(GEN x, GEN y, long l)
{
      /* a usage interne. x,y sont de type 6 */
  GEN z;

  z=gsub((GEN)x[1],(GEN)y[1]);
  if(gcmp0(z)||(gexpo(z)<(2-((l-2)<<TWOPOTBITS_IN_LONG))))
    {
      z=gadd((GEN)x[2],(GEN)y[2]);
      return gcmp0(z)||(gexpo(z)<(2-((l-2)<<TWOPOTBITS_IN_LONG)));
    }
  else return 0;
}
  
GEN 
roots(GEN p, long l)
{
  long av,tetpil,n,j,k,s,fl;
  GEN p1,p2,p3,p4,p22,p33;
  
  av=avma;p1=roots_com(p,l);n=lg(p1);
  if(isrealappr(p,l))
    {
      p3=cgetg(n,18);s=0;
      for(j=1;j<n;j++)
	{
	  p2=(GEN)p1[j];
	  if(typ(p2)<6) {p3[++s]=(long)p2;p1[j]=zero;}
	  else if(isrealappr((GEN)p1[j],l)) {p3[++s]=p2[1];p1[j]=zero;}
	}
      setlg(p3,s+1);p4=sort(p3);setlg(p3,n);
      for(j=1;j<=s;j++)
	{
	  p2=cgetg(3,6);p2[1]=lgetr(l);p2[2]=lgetr(l);
	  gaffect((GEN)p4[j],(GEN)p2[1]);affsr(0,(GEN)p2[2]);p3[j]=(long)p2;
	}
      for(j=1;j<n;j++)
	{
	  p2=(GEN)p1[j];
	  if(typ(p2)>=6)
	    {
	      p3[++s]=(long)tocomplex(p2,l);p33=(GEN)p3[s];
	      fl=0;
	      for(k=j+1;(k<n)&&(!fl);k++)
		{
		  p22=(GEN)p1[k];
		  if(typ(p22)>=6)
		    {
		      fl=isconj(p33,p22,l);
		      if(fl) {p3[++s]=(long)tocomplex(p22,l);p1[k]=zero;}
		    }
		}
	      if(!fl) err(talker,"bug1 conjugates in roots");
	    }
	}
      if(s!=(n-1)) err(talker,"bug2 conjugates in roots");
      tetpil=avma;return gerepile(av,tetpil,gcopy(p3));
    }
  else
    {
      tetpil=avma;p2=cgetg(n,18);
      for(j=1;j<n;j++) p2[j]=(long)tocomplex((GEN)p1[j],l);
      return gerepile(av,tetpil,p2);
    }
}

